from speedrunapi import Game
from speedrunapi.errors import DataFetchError

class Categories:
    
    @property
    def categories(self):
        """Returns all of the category data avalible"""
        return self.game_category_data
    
    def game_categories_data(self, keys):
        final_data = []
        for data in range(len(self.game_category_data)):
            final_data.append(self.game_category_data[data][keys])
        return final_data
    
    @property
    def category_ids(self):
        """Returns a list of the ids for categories of a game"""
        return self.game_categories_data("id")
    
    @property
    def category_names(self):
        """Returns the names of the categroies"""
        return self.game_categories_data("name")
    
    @property
    def category_types(self):
        return self.game_categories_data("type")
    
    @property
    def category_rules(self):
        """Returns a list of the rules associated with each category"""
        final_rules = []
        rules = self.game_categories_data("rules")
        for rule in range(len(rules)):
            final_rules.append(rules[rule].replace('\n','').replace('\r',''))
        return final_rules
    
    @property
    def category_players(self):
        return self.game_categories_data("players")
    
    @property
    def category_urls(self):
        """Returns a list of the urls related to the category.
        
        In order the return is
        [variable_url, record_url, runs_url, leaderboards_url]"""
        self.final_urls = []
        urls = ['variables', 'records', 'runs', 'leaderboard']
        data = self.game_categories_data("links")
        try:
            for url in range(len(urls)):
                self.final_urls.append(data[url][url]['uri'])
            return self.final_urls
        except IndexError:
            raise DataFetchError(f"Incomplete URl set for this category. \nUrls found:{self.final_urls}")

    def __init__(self, game):
        """Category data for speedrun.com
        
        ARGS:
        game (str): The name of the game you want to get the category data for"""
        self.game = game
        self.game_data = Game(self.game)
        self.game_category_data = self.game_data.catagories
        


"""
A python library for working with the speedrun.com API.
This library is intended to make the api easier to use,
by preprocessing the data, and separating it into specific sub categories.
"""

from .speedrunapi import *
from speedrunapi.game_data import *
from speedrunapi.user_data import *
from speedrunapi.runs import *
from speedrunapi.levels import *
from speedrunapi.regions import * 
from speedrunapi.leaderboard import *
from speedrunapi.categories import *



from speedrunapi import User_Requests, Game_Requests, Game
from datetime import datetime
# add a way to set the catagories for game run

class User_Runs:
    """Data about a users runs on speedrun.com
   
    ARGS: 
    User - The user you want to get the runs for"""

    @property
    def runs(self):
        """Returns a dictionary of all the runs a user has on speedrun.com"""
        return self.user_run_data

    def user_runs_data(self, keys):
        final_run_data = []
        for run in range(len(self.user_run_data)):
            final_run_data.append(self.user_run_data[run][keys])
        return final_run_data

    @property
    def run_ids(self):
        """Returns a list of all the ids of runs a user has on speedrun.com"""
        return self.user_runs_data("id")

    @property
    def run_games(self):
        """Returns a list of all of the games a user has run on speedrun.com"""
        return self.user_runs_data("game")

    @property
    def run_levels(self):
        """Returns a list of all of the levels a user has run on speedrun.com"""
        return self.user_runs_data("level")

    @property
    def run_catagories(self):
        """Returns a list of all of the catagories a user has run on speedrun.com"""
        return self.user_runs_data("category")

    @property
    def run_videos(self):
        """Returns a list of all of the videos a user has for runs on speedrun.com"""
        videos = []
        for video in range(len(self.user_run_data)):
            try:
                videos.append(self.user_run_data[video]["videos"]["links"][0]["uri"])
            except TypeError:
                continue
            except IndexError:
                break
        return videos

    @property
    def run_comments(self):
        """Returns a list of all of the comments a user has for runs on speedrun.com"""
        return self.user_runs_data("comment")

    @property
    def run_status(self):
        """Returns a list of the status of a users runs on speedrun.com"""
        status = []
        for stat in range(len(self.user_run_data)):
            status.append(self.user_run_data[stat]["status"]["status"])
        return status

    @property
    def run_examiners(self):
        """Returns a list of the users who have examined a users runs on speedrun.com"""
        examiners = []
        for examiner in range(len(self.user_run_data)):
                examiners.append(self.user_run_data[examiner]["status"]["examiner"])
        return examiners

    @property
    def run_verify_dates(self):
        """Returns a list of the dates when a users runs were verified on speedrun.com"""
        verify_dates = []
        for verify_date in range(len(self.user_run_data)):
            unformatted_date = self.user_run_data[verify_date]["status"]["verify-date"]
            formatted_date = datetime.fromisoformat(unformatted_date)
            verify_dates.append(formatted_date.strftime("%Y-%m-%d %H:%M:%S"))
        return verify_dates

    @property
    def run_submitted_dates(self):
        """Returns a list of the dates when a users runs were submitted on speedrun.com"""
        submitted_dates = []
        for submitted_date in range(len(self.user_run_data)):
            unformatted_date = self.user_run_data[submitted_date]["submitted"]
            formatted_date = datetime.fromisoformat(unformatted_date)
            submitted_dates.append(formatted_date.strftime("%Y-%m-%d %H:%M:%S"))
        return submitted_dates
    
    @property
    def run_times(self):
        run_times = []
        for time in range(len(self.user_run_data)):
            unformatted_date = self.user_run_data[time]["times"]["primary_t"]
            if unformatted_date / 60 < 1:
                seconds = "S:", unformatted_date
                run_times.append(seconds)
            elif unformatted_date / 3600 < 1:
                minutes = "M:", round(unformatted_date / 60, 2)
                run_times.append(minutes)
            else:
                hours = "H:", round(unformatted_date / 3600, 2)
                run_times.append(hours)
        return run_times

    def __init__(self, user):
        self.user = user
        self.user_data = User_Requests(self.user)
        self.user_run_data = self.user_data.user_data_request(1)


class Game_Runs:
    """Data about a certin games runs on speedrun.com
    
    ARGS: 
    Game - The game you want to get the runs for either as an id or name.
    (SOON!) Catagory ID/Name - the ID of the specific catagory you want to get runs for."""

    def game_runs_data(self, keys):
        self.final_run_data = []
        for run in range(len(self.game_run_data)):
            self.final_run_data.append(self.game_run_data[run][keys])
        return self.final_run_data

    @property
    def runs(self):
        """Returns a dictionary of all the runs a user has on speedrun.com"""
        return self.game_run_data

    @property
    def run_ids(self):
        """Returns a list of all the ids of runs a user has on speedrun.com"""
        return self.game_runs_data("id")

    @property
    def run_levels(self):
        """Returns a list of all of the levels a user has run on speedrun.com"""
        return self.game_runs_data("level")

    @property
    def run_catagories(self):
        """Returns a list of all of the catagories a user has run on speedrun.com"""
        return self.game_runs_data("category")

    @property
    def run_videos(self):
        """Returns a list of all of the videos a game has for runs on speedrun.com"""
        videos = []
        for video in range(len(self.game_run_data)):
            try:
                videos.append(self.game_run_data[video]["videos"]["links"][0]["uri"])
            except TypeError:
                continue
            except IndexError:
                break
        return videos

    @property
    def run_comments(self):
        """Returns a list of all of the comments a game has for runs on speedrun.com"""
        return self.game_runs_data("comment")

    @property
    def run_status(self):
        """Returns a list of the status of a users runs on speedrun.com"""
        status = []
        for stat in range(len(self.game_run_data)):
            status.append(self.game_run_data[stat]["status"]["status"])
        return status

    @property
    def run_examiners(self):
        """Returns a list of the users who have examined a users runs on speedrun.com"""
        examiners = []
        for examiner in range(len(self.game_run_data)):
            examiners.append(self.game_run_data[examiner]["status"]["examiner"])
        return examiners

    @property
    def run_verify_dates(self):
        """Returns a list of the dates when a users runs were verified on speedrun.com"""
        verify_dates = []
        for verify_date in range(len(self.game_run_data)):
            try:
                unformatted_date = self.game_run_data[verify_date]["status"]["verify-date"]
                if unformatted_date == None:
                    verify_dates.append(None)
                    continue
                formatted_date = datetime.fromisoformat(unformatted_date)
                verify_dates.append(formatted_date.strftime("%Y-%m-%d %H:%M:%S"))
            except KeyError:
                break
        return verify_dates

    @property
    def run_users(self):
        """Returns a list of the users who have submitted a run on speedrun.com"""
        users = []
        for user in range(len(self.game_run_data)):
            if self.game_run_data[user]["players"][0]["rel"] == "guest":
                users.append(self.game_run_data[user]["players"][0]["name"])
            else:
                users.append(self.game_run_data[user]["players"][0]["id"])
        return users

    @property
    def run_submitted_dates(self):
        """Returns a list of the dates when a users runs were submitted on speedrun.com"""
        submitted_dates = []
        for submit_date in range(len(self.game_run_data)):
            unformatted_date = self.game_run_data[submit_date]["submitted"]
            if unformatted_date == None:
                submitted_dates.append(None)
                continue
            formatted_date = datetime.fromisoformat(unformatted_date)
            submitted_dates.append(formatted_date.strftime("%Y-%m-%d %H:%M:%S"))
        return submitted_dates

    @property
    def run_times(self):
        """Returns a list of the times for runs on speedrun.com"""
        run_times = []
        for time in range(len(self.game_run_data)):
            unformatted_date = self.game_run_data[time]["times"]["primary_t"]
            if unformatted_date / 60 < 1:
                seconds = "S:", unformatted_date
                run_times.append(seconds)
            elif unformatted_date / 3600 < 1:
                minutes = "M:", round(unformatted_date / 60, 2)
                run_times.append(minutes)
            else:
                hours = "H:", round(unformatted_date / 3600, 2)
                run_times.append(hours)
        return run_times

    @property
    def run_platforms(self):
        """Returns a list of the platforms for runs on speedrun.com"""
        platforms = []
        for platform in range(len(self.game_run_data)):
                platforms.append(self.game_run_data[platform]["system"]["platform"])
        return platforms

    @property
    def run_values(self):
        """Returns a list of the values for runs on speedrun.com"""
        values = []
        for value in range(len(self.game_run_data)):
            values.append(self.game_run_data[value]["values"])
        return values

    def __init__(self, game, category = None):
        self.game = game
        self.game_data = Game(self.game)
        self.game_id = self.game_data.id
        self.game_data = Game_Requests(self.game_id)
        self.game_data.game_request()
        self.game_run_data = self.game_data.game_data_request(1)



from speedrunapi import User_Requests
from datetime import datetime
import http.client
from .errors import URLerror

class User:
    """Data about a user on speedrun.com
    ARGS: user, a username of a person on speedrun.com"""

    @property
    def lookup(self):
        """Returns the users full_data on speedrun.com"""
        return self.user_data.user_data

    @property
    def id(self):
        """Returns the id of the user as a string"""
        userid = [id.get("id") for id in self.user_data.user_data]
        return userid[0]

    @property
    def name_style(self):
        """Returns the name style of the user as a dictionary"""
        return self.user_data.user_data[0]["name-style"]

    @property
    def role(self):
        """Returns the speedrun.com role of the user as a string"""
        return self.user_data.user_data[0]["role"]

    @property
    def signup(self):
        """Returns the date the user joined speedrun.com as a string in the format Y-M-D H:M:S"""
        date_unformatted = self.user_data.user_data[0]["signup"]
        date_formatted = datetime.fromisoformat(date_unformatted)
        return date_formatted.strftime("%Y-%m-%d %H:%M:%S")

    @property
    def location(self):
        """Returns the location the user has listed on speedrun.com as a dictionary"""
        return self.user_data.user_data[0]["location"]

    @property
    def links(self):
        """Returns the links the user has listed on their profile on speedrun.com as a list"""
        final_links = []
        link_types = ["twitch", "hitbox", "youtube", "twitter", "speedunslive"]
        for link_type in range(len(link_types)):
            final_links.append([link.get(link_types[link_type]) for link in self.user_data.user_data])
        return final_links

    # special cases

    @property
    def moderated_games(self):
        """Returns a dictionary of all the moderated games a user has on speedrun.com"""
        self.user_data.user_data_request(2)
        moderated = self.user_data.user_data_response
        if len(moderated) == 0:
            return "User moderates no games"
        else:
            return moderated

    @property
    def personal_bests(self):
        """Returns a dictionary of the personal bests of a user on speedrun.com"""
        self.user_data.user_data_request(3)
        return self.user_data.user_data_response

    def __init__(self, user):
        try:
            self.user = user
            self.user_data = User_Requests(self.user)
        except http.client.InvalidURL as e:
            details = e.args[0].split("'")[2]
            raise URLerror(f'{details}\nCheck to make sure the name acctualy has no alphanumeric characters/spaces')



from json import loads
from urllib.request import urlopen
from functools import lru_cache
import urllib.error
from speedrunapi.errors import URLerror
#start doing error handleing here

class User_Requests:
    """Makes the user requests to speedrun.com"""
    
    @lru_cache(maxsize = 10)  
    def user_reqest(self):
        """Makes a request to speedrun.com"""
        url = f'https://www.speedrun.com/api/v1/users?lookup={self.user}'
        responce = urlopen(url)
        self.user_data = loads(responce.read().decode('utf-8'))['data']
        return self.user_data

    @lru_cache(maxsize = 10)
    def user_data_request(self, x):
        """Makes a request to speedrun.com to get the full data of the user
        
        ARGS:
        x (int): The index of the link being retrieved"""
        self.user_stat_links()
        url = self.final_links[x]
        responce = urlopen(url)
        self.user_data_response = loads(responce.read().decode('utf-8'))['data']
        return self.user_data_response
            
    @lru_cache(maxsize = 10)
    def user_stat_links(self):
        """Returns the links for the users other pages such as their runs,personal bests, etc. As a list"""
        self.final_links = []
        link_types = ['self', 'runs', 'games', 'personal-bests']
        for link_type in range(len(link_types)):
            self.final_links.append(self.user_data[0]['links'][link_type]['uri'])
        return self.final_links
    
    def __init__(self, user):
        self.user = user
        self.user_reqest()
        
class Game_Requests:
    """Makes the game requests to speedrun.com"""
    
    @lru_cache(maxsize = 10)  
    def game_reqest_quick(self):
        """Makes a request to speedrun.com just to get the nessiisary info to do a better query"""
        url = f'https://www.speedrun.com/api/v1/games?name={self.game}'
        response = urlopen(url)
        self.quick_game_data = loads(response.read().decode('utf-8'))
        return self.quick_game_data
    
    @lru_cache(maxsize = 10)
    def game_request(self):
        """Makes a request to speedrun.com to get the nessisary info to do a better query"""
        url = f'https://www.speedrun.com/api/v1/games/{self.game}'
        responce = urlopen(url)
        self.game_data = loads(responce.read().decode('utf-8'))['data']
        return self.game_data
    
    @lru_cache(maxsize = 10)
    def game_data_request(self, x):
        """Uses the links in user_stat_links to retrieve different data for more specific topics
        ARGS:
        x (int): The index of the link in the user_stat_links list"""
        self.game_stat_links()
        url = self.final_links[x]
        responce = urlopen(url)
        self.game_data_response = loads(responce.read().decode('utf-8'))['data']
        return self.game_data_response
    
    @lru_cache(maxsize = 10)
    def game_stat_links(self):
        """Returns the links for the runs, levels, etc. for the game as a list"""
        self.final_links = []
        link_types = ['self', 'runs', 'levels', 'catagories', 'variables', 'records', 'series', 'base-game', 'derived-games', 'romhacks', 'leaderboard']
        for link_type in range(len(link_types)-1):
            self.final_links.append(self.game_data['links'][link_type]['uri'])
        return self.final_links

    def __init__(self, game):
        self.game = game

class Misc_Requests:

    @lru_cache(maxsize = 10)
    def region_data_request(url):
        """Returns the data for regions after provideing a url to get it"""
        response = urlopen(url) #type: ignore python cant see the url supplied as a parmater
        region_data = loads(response.read().decode('utf-8'))['data']
        return region_data

    @lru_cache(maxsize = 10)
    def region_link_request(region, x):
        """Gets the links for the region data
        ARGS:
        x (int): The index of the link to get"""
        region = region.split(' ', 1)[0] #type: ignore ,error with python thinking its a callable value when its not
        try:
            url = f'https://www.speedrun.com/api/v1/regions/{region}'
        except urllib.error.HTTPError:
            try: 
                url = f'https://www.speedrun.com/api/v1/games?region={region}'
            except urllib.error.HTTPError:
                raise URLerror("Invalid region")
        responce = urlopen(url)
        region_data = loads(responce.read().decode('utf-8'))['data']['links'][x]['uri']
        return region_data
    
    @lru_cache(maxsize = 10)
    def leaderboard_data_request(url):
        """Returns data for leaderboards on speedrun.com"""
        try:
            response = urlopen(url) #type: ignore
            leaderboard_data = loads(response.read().decode('utf-8'))['data']
        except urllib.error.HTTPError:
            raise URLerror("Invalid game or category")
        return leaderboard_data



from speedrunapi import Misc_Requests
from speedrunapi.errors import DataFetchError
from speedrunapi.translation import *
from datetime import datetime
import asyncio
from aiohttp import ClientSession
from functools import lru_cache

class Leaderboard:
    @lru_cache(maxsize=100)
    async def translate_user_id(self,):  # this allows for large data requests to be made faster internaly for the library for matters like translation
        async with ClientSession() as session:
            self.names = []
            for name in self.run_users:
                try:
                    url = f"https://www.speedrun.com/api/v1/users/{name}"
                    async with session.get(url) as resp:
                        name = await resp.json()
                        self.names.append(name["data"]["names"]["international"])
                except KeyError:
                    pass
            return self.names

    @property
    def leaderboard(self):
        """All the data for a leaderboard on speedrun.com"""
        return self.leaderboard_data

    @property
    def leaderboard_top(self):
        """Returns the leaderboard with the person and place (can cause lots of lag)
        Takes in the argument of the top amount of places you want (defaults to 50000 or whatever the max is)"""
        final_places = []
        if self.limit is None:
            self.limit = 50000
        for place in range(0, self.limit):
            try:
                place = self.leaderboard_data["runs"][place]["place"]
                if self.leaderboard_data["runs"][place]["run"]["players"][0]["rel"] == "guest":
                    final_places.append(None)
                name = self.leaderboard_data["runs"][place]["run"]["players"][0]["id"]
                final_places.append(str(place) + " : " + str(translate_user_id(name)))
            except KeyError:
                final_places.append(None)
            except IndexError:
                break
        return final_places

    @property
    def leaderboard_run_ids(self):
        """Returns a list of the ids repersenting each indivdual run for the leaderboard"""
        if self.limit is None:
            self.limit = len(self.leaderboard_data["runs"])
        run_ids = []
        for run_id in range(0, self.limit):
            try:
                run_ids.append(self.leaderboard_data["runs"][run_id]["run"]["id"])
            except IndexError:
                break
        return run_ids

    @property
    def leaderboard_run_comments(self):
        """Returns a list of the comments for each of the runs on the leaderboard"""
        if self.limit is None:
            self.limit = len(self.leaderboard_data["runs"])
        run_comments = []
        for run_id in range(0, self.limit):
            try:
                comment = self.leaderboard_data["runs"][run_id]["run"]["comment"]
                if str(comment).__contains__("\n") or str(comment).__contains__("\r"):
                    comment = comment.replace("\n", "").replace("\r", "")
                run_comments.append(comment)
            except IndexError:
                break
        return run_comments

    @property
    def leaderboard_run_status(self) -> list:
        """Returns a list of the status of runs.
        The values you will see in each list index are:
        status: verified - meaning that the run is verified to be not cheated,
        examiner: (name) - the name of the person that verified the run,
        verify-date: (date) - the date the run was verified."""
        if self.limit is None:
            self.limit = len(self.leaderboard_data["runs"])
        run_status = []
        for run_id in range(0, self.limit):
            try:
                status = self.leaderboard_data["runs"][run_id]["run"]["status"]
                status["examiner"] = translate_user_id(status["examiner"])
                if status["status"] == "rejected":
                    pass
                elif status["verify-date"]:
                    date_formatted = datetime.fromisoformat(status["verify-date"])
                    status["verify-date"] = date_formatted.strftime("%Y-%m-%d %H:%M:%S")
                run_status.append(status)
            except IndexError:
                break
        return run_status

    def leaderboard_run_users(self, translation=False):
        """Returns the users who have made a run on speedrun.com in order.
        CAN TAKE A LONG TIME DEPENDING ON THE SIZE OF THE LEADERBOARD

        ARGS:
        translation (bool): if True, translate the user_id to their international name,
        if False (default), just return the user_ids for improved performance"""
        if self.limit is None:
            self.limit = len(self.leaderboard_data["runs"])
        self.run_users = []
        for run_id in range(0, self.limit):
            try:
                user = self.leaderboard_data["runs"][run_id]["run"]["players"]
                if user[0]["rel"] == "guest":
                    user = user[0]["name"]
                else:
                    user = user[0]["id"]
                self.run_users.append(user)
            except IndexError:
                break
        if translation:
            people = asyncio.run(self.translate_user_id())
            return people
        else:
            return self.run_users

    @property
    @lru_cache(maxsize=100)
    def leaderboard_run_times(self):
        """Returns a list of the times of each run on the leaderboard in order
        What you should see:
        'H', 'M', or 'S' meaning hours, minutes, or seconds respecivly.
        A float with two decimal places repersenting the time.

        Examples:
        - ('H': 2.42) repersenting a time of 2 hours and 42 minutes.
        - ('M': 2.42) repersenting a time of 2 minutes and 42 seconds.
        - ('S': 2.42) repersenting a time of 2 seconds and 42 miliseconds."""
        if self.limit is None:
            self.limit = len(self.leaderboard_data["runs"])
        run_times = []
        for run_time in range(0, self.limit):
            try:
                unformatted_date = self.leaderboard_data["runs"][run_time]["run"]["times"]["primary_t"]
                if unformatted_date / 60 < 1:
                    seconds = "S:", unformatted_date
                    run_times.append(seconds)
                elif unformatted_date / 3600 < 1:
                    minutes = "M:", round(unformatted_date / 60, 2)
                    run_times.append(minutes)
                else:
                    hours = "H:", round(unformatted_date / 3600, 2)
                    run_times.append(hours)
            except IndexError:
                break
        return run_times

    @property
    @lru_cache(maxsize=100)
    def leaderboard_run_platforms(self):
        """Returns a list of the platforms used for each of the runs in order."""
        if self.limit is None:
            self.limit = len(self.leaderboard_data["runs"])
        run_platforms = []
        for run_platform in range(0, self.limit):
            try:
                platform = self.leaderboard_data["runs"][run_platform]["run"]["system"]["platform"]
                run_platforms.append(translate_platform(platform))
            except IndexError:
                break
        return run_platforms

    @property
    @lru_cache(maxsize=100)
    def leaderboard_run_regions(self):
        """Returns a list of the regions each of the runs was  in."""
        if self.limit is None:
            self.limit = len(self.leaderboard_data["runs"])
        run_regions = []
        for run_region in range(0, self.limit):
            try:
                region = self.leaderboard_data["runs"][run_region]["run"]["system"]["region"]
                if region:
                    run_regions.append(translate_region(region))
                else:
                    run_regions.append(region)
            except IndexError:
                break
        return run_regions

    # TODO ADD TRANSLATION FOR RUN VALUES AND FIGURE OUT WHAT THEY ARE

    def __init__(self, game, category=None, limit=None):
        """LEADERBOARD data from the speedrun.com API

        Args:
            game (str): The name of the game you want to get the leaderboard for
            category (str, optional): The category you want to get the leaderboard for.
        If not provided will take the defult one from speedrun.com
            limit (int, optional): The amount of runs you want to have returned
        Defaults to the maximum amount"""
        from speedrunapi import Categories

        self.limit = limit
        # checks to see if you have provided the category, if you have not then it takes the default
        if category:
            url = f"https://www.speedrun.com/api/v1/leaderboards/{game_name_to_id(game)}/category/{category_name_to_id(game, category)}"
        else:
            category = Categories(game)
            try:
                url = category.category_urls[3]
            except DataFetchError:
                url = category.final_urls[0]
                url += "/records"
        self.leaderboard_data = Misc_Requests.leaderboard_data_request(url)



from json import loads
from urllib.request import urlopen
import urllib.error
import http.client
from functools import lru_cache
from speedrunapi.errors import URLerror
#try to make just one request or just a couple of requests

@lru_cache()
def username_to_id(name):
    """Turns a username of a player on speedrun.com into the id of that player"""
    try:
        url = f'https://www.speedrun.com/api/v1/users?name={name}'
        return loads(urlopen(url).read().decode('utf-8'))['data'][0]['id']
    except IndexError:
        return "No name found"
    except http.client.InvalidURL as e:
        details = e.args[0].split("'")[2]
        raise URLerror(f'{details}\nNames cannot contain a non alphanumeric character')

@lru_cache()    
def game_name_to_id(name):
    """Turns a game name of a player on speedrun.com into the id of that game"""
    try:
        name = name.replace(' ','_')
        url = f'https://www.speedrun.com/api/v1/games?name={name}'
        return loads(urlopen(url).read().decode('utf-8'))['data'][0]['id']
    except IndexError:
        return "No game found"

@lru_cache()   
def game_name_to_abbreviation(name):
    """Turns a game name of a player on speedrun.com into the abbreviation of that game"""
    try:
        name=name.replace(' ','_')
        url = f'https://www.speedrun.com/api/v1/games?name={name}'
        return loads(urlopen(url).read().decode('utf-8'))['data'][0]['abbreviation']
    except IndexError:
        return "No game found"

@lru_cache()
def category_name_to_id(game, category_name):
    """Turns a name of a category into an id
    
    ARGS:
        game (str): The name of the game
        category_name (str): The name of the category"""
    from speedrunapi import Categories
    try:
        category_data = Categories(game)
        for category in range(len(category_data.category_names)):
            if category_data.category_names[category].lower() == category_name.lower():
                return category_data.category_ids[category]
    except IndexError:
        return "No category found"

@lru_cache()
def translate_user_id(userid):
    """Turns a user id of a user into the username of the user"""
    try:
        url = f'https://www.speedrun.com/api/v1/users/{userid}'
        return loads(urlopen(url).read().decode('utf-8'))['data']['names']['international'] 
    except urllib.error.HTTPError:
        return "No user found"
    except http.client.InvalidURL as e:
        details = e.args[0].split("'")[2]
        raise URLerror(f'{details}\nIDs cannot contain a non alphanumeric character')

@lru_cache()  
def translate_game_id(gameid):
    """Turns a game id of a game into the name of the game"""
    try:
        url = f'https://www.speedrun.com/api/v1/games/{gameid}'
        return loads(urlopen(url).read().decode('utf-8'))['data']['names']['international'] 
    except urllib.error.HTTPError:
        return "No game found"
    except http.client.InvalidURL as e:
        details = e.args[0].split("'")[2]
        raise URLerror(f'{details}\nIDs cannot contain a non alphanumeric character')

@lru_cache()   
def translate_game_abbreviation(game_abbreviation):
    """Turns the game abbreviation of a game into the name of the game"""
    try:
        url = f'https://www.speedrun.com/api/v1/games/{game_abbreviation}'
        return loads(urlopen(url).read().decode('utf-8'))['data']['names']['international'] 
    except urllib.error.HTTPError:
        return "No game found"
    except http.client.InvalidURL as e:
        details = e.args[0].split("'")[2]
        raise URLerror(f'{details}\nAbbreviations cannot contain a non alphanumeric character')

@lru_cache()   
def translate_region(regionid):
    """Turns the region id of a region into the name of the region
    
    ARGS:
        regionid (str): The id of the region"""
    try:
        url = f'https://www.speedrun.com/api/v1/regions/{regionid}'
        return loads(urlopen(url).read().decode('utf-8'))['data']['name']
    except urllib.error.HTTPError:
        return "No region found"
    except http.client.InvalidURL as e:
        details = e.args[0].split("'")[2]
        raise URLerror(f'{details}\nregion IDs cannot contain a non alphanumeric character')

@lru_cache()    
def translate_platform(platformid):
    """Turns the platform id of a platform into the name of the platform"""
    try:
        url = f'https://www.speedrun.com/api/v1/platforms/{platformid}'
        return loads(urlopen(url).read().decode('utf-8'))['data']['name']
    except urllib.error.HTTPError:
        return "No platform found"
    except http.client.InvalidURL as e:
        details = e.args[0].split("'")[2]
        raise URLerror(f'{details}\nplatform IDs cannot contain a non alphanumeric character')

@lru_cache()   
def translate_genre(genreid):
    """Turns the genre id of a genre into the name of the genre"""
    try:
        url = f'https://www.speedrun.com/api/v1/genres/{genreid}'
        return loads(urlopen(url).read().decode('utf-8'))['data']['name']
    except urllib.error.HTTPError:
        return "No genre found"
    except http.client.InvalidURL as e:
        details = e.args[0].split("'")[2]
        raise URLerror(f'{details}\ngenre IDs cannot contain a non alphanumeric character')

@lru_cache()   
def translate_engine(engineid):
    """Turns the engine id of a engine into the name of the engine"""
    try:
        url = f'https://www.speedrun.com/api/v1/engines/{engineid}'
        return loads(urlopen(url).read().decode('utf-8'))['data']['name']
    except urllib.error.HTTPError:
        return "No engine found"
    except http.client.InvalidURL as e:
        details = e.args[0].split("'")[2]
        raise URLerror(f'{details}\nengine IDs cannot contain a non alphanumeric character')

@lru_cache()
def translate_level(levelid):
    """Turns the level id of a level into the name of the level"""
    try:
        url = f'https://www.speedrun.com/api/v1/levels/{levelid}'
        return loads(urlopen(url).read().decode('utf-8'))['data']['name']
    except urllib.error.HTTPError:
        return "No level found"
    except http.client.InvalidURL as e:
        details = e.args[0].split("'")[2]
        raise URLerror(f'{details}\nlevel IDs cannot contain a non alphanumeric character')

@lru_cache()
def translate_category(categoryid):
    """Turns the category id of a category into the name of the category
    
    ARGS:
    categoryID: the id of the category"""
    try:
        url = f'https://www.speedrun.com/api/v1/categories/{categoryid}'
        return loads(urlopen(url).read().decode('utf-8'))['data']['name']
    except urllib.error.HTTPError:
        return "No category found"
    except http.client.InvalidURL as e:
        details = e.args[0].split("'")[2]
        raise URLerror(f'{details}\ncategory IDs cannot contain a non alphanumeric character')

@lru_cache()
def translate_variable(variableid):
    """Turns the variable id of a variable into the name of the variable"""
    try:
        url = f'https://www.speedrun.com/api/v1/variables/{variableid}'
        return loads(urlopen(url).read().decode('utf-8'))['data']['name']
    except urllib.error.HTTPError:
        return "No variable found"
    except http.client.InvalidURL as e:
        details = e.args[0].split("'")[2]
        raise URLerror(f'{details}\nvariable IDs cannot contain a non alphanumeric character')

@lru_cache()   
def translate_level_id(levelid):
    """Turns the level id of a level into the name of the level"""
    try:
        url = f'https://www.speedrun.com/api/v1/levels/{levelid}'
        return loads(urlopen(url).read().decode('utf-8'))['data']['name']
    except urllib.error.HTTPError:
        return "No level found"
    except http.client.InvalidURL as e:
        details = e.args[0].split("'")[2]
        raise URLerror(f'{details}\nlevel IDs cannot contain a non alphanumeric character')



from speedrunapi import Game, Game_Requests

class Levels:
    
    def game_levels_data(self, keys):
        self.final_level_data = []
        for level in range(0, 100000):
            try:
                self.final_level_data.append(self.game_level_data[level][keys])# type: ignore            
            except IndexError:
                break
        return self.final_level_data

    @property
    def levels(self):
        return self.game_level_data
    
    @property
    def level_ids(self):
        return self.game_levels_data("id")
    
    @property
    def level_names(self):
        return self.game_levels_data("name")
    
    @property
    def level_rules(self):
        return self.game_levels_data("rules")
    
    def __init__(self, game):
        self.game = game
        self.game_data = Game(self.game)
        self.game_id = self.game_data.id
        self.game_data = Game_Requests(self.game_id)
        self.game_data.game_request()
        self.game_level_data = self.game_data.game_data_request(2)



"""Custom errors for the library."""

class URLerror(Exception):
    def __init__(self, message):
        """This error is thrown when the url is invalid.
        This replaces http.client.InvalidURL errors."""
        super().__init__(message)

class Region_Length_Error(Exception):
    def __init__(self, message):
        """This error is thrown when there is more then 1 region
        provided when calling the region module. Or there is 0
        regions provided"""
        super().__init__(message)

class InvalidRegionError(Exception):
    def __init__(self, message):
        """This error is thrown when the region provided is invalid."""
        super().__init__(message)

class DataFetchError(Exception):
    def __init__(self, message):
        """This error is called when there was an error fetching all the data."""
        super().__init__(message)


from speedrunapi import Misc_Requests
from speedrunapi.errors import Region_Length_Error, InvalidRegionError
from speedrunapi.translation import * 
from datetime import datetime
import urllib.error
# !region variables?

class Regions:
    """Data about regions on speedrun.com
    
    ARGS: 
    region - The region id or name of a region on speedrun.com"""
    
    @property
    def region_game(self) -> dict:
        """Gets all the data about games inside the region"""
        try:
            link = Misc_Requests.region_link_request(self.region[0], 1)
        except TypeError:
            link = Misc_Requests.region_link_request(self.region[0][0], 1)
        except urllib.error.HTTPError:
            raise InvalidRegionError(f'"{self.region[0]}" The region provided is not a valid region, check your spelling.')
        return Misc_Requests.region_data_request(link)
    
    @property
    def region_game_names(self) -> list:
        """Returns a list of the names of the games that are included in the region"""
        final_game_names = []
        for id in range(len(self.region_runs)):
            final_game_names.append(self.region_game[id]['names']['international'])
        return final_game_names
    
    @property
    def region_runs(self) -> dict:
        """Returns all of the information avalible for runs in a region"""
        try:
            link = Misc_Requests.region_link_request(self.region[0], 2)
        except TypeError:
            link = Misc_Requests.region_link_request(self.region[0][0], 2)
        except urllib.error.HTTPError:
            raise InvalidRegionError(f'"{self.region[0]}" The region provided is not a valid region, check your spelling.')
        return Misc_Requests.region_data_request(link)
    
    @property
    def region_run_ids(self):
        """Returns a list of the ids for each of the runs in the region"""
        final_runs_ids = []
        for id in range(len(self.region_runs)):
            final_runs_ids.append(self.region_runs[id]['id'])
        return final_runs_ids
    
    @property
    def region_run_games(self):
        """Returns a list of the games that are in the reigon"""
        final_games = []
        for id in range(len(self.region_runs)):
            game_id = self.region_runs[id]['game']
            final_games.append(translate_game_id(game_id))
        return final_games
    
    @property
    def region_run_levels(self):
        """Returns the levels run inside the region"""
        final_levels = []
        for id in range(len(self.region_runs)):
                level = self.region_runs[id]['level']
                final_levels.append(translate_level(level))
        return final_levels
    
    @property
    def region_run_categories(self):
        """Returns the categories for the runs inside of the region"""
        final_categories = []
        for id in range(len(self.region_runs)):
                catagory = self.region_runs[id]['category']
                final_categories.append(translate_catagory(catagory))
        return final_categories
    
    @property
    def region_run_status(self):
        """Returns the status's of each of the runs in the region
        
        Format:
        'status': Can either be verified or rejected showing if the run was allowed on the leaderboards
        'examiner': The name of the user who verified the run
        'verify-date': The date when the run was verified"""
        final_status = []
        for id in range(len(self.region_runs)):
            status = self.region_runs[id]['status']
            status['examiner'] = translate_user_id(status['examiner'])
            if status['status'] == 'rejected':
                pass
            elif status['verify-date']:
                date_formatted = datetime.fromisoformat(status['verify-date'])
                status['verify-date'] = date_formatted.strftime("%Y-%m-%d %H:%M:%S")
            final_status.append(status)
        return final_status
    
    @property
    def region_run_comments(self):
        """Returns a list of the comments that came with the region"""
        final_comments = []
        for id in range(len(self.region_runs)):
            comment = self.region_runs[id]['comment']
            if str(comment).__contains__('\n') or str(comment).__contains__('\r'):
                comment = comment.replace('\n','').replace('\r','')
            final_comments.append(comment)
        return final_comments
    
    @property
    def region_run_users(self):
        """Reurns a list of the names of players who have made runs inside the region"""
        final_users = []
        for id in range(len(self.region_runs)):
            user = self.region_runs[id]['players']
            if user[0]['rel'] == 'guest':
                user = user[0]['name']
            else:
                user = translate_user_id(user[0]['id'])
            final_users.append(user)
        return final_users
    
    @property
    def region_run_submitted_dates(self):
        """Returns the dates runs were subbmitted on"""
        final_date = []
        for id in range(len(self.region_runs)):
            date = self.region_runs[id]['date']
            date_formatted = datetime.fromisoformat(date)
            final_date.append(date_formatted.strftime("%Y-%m-%d"))
        return final_date
    
    @property
    def region_run_times(self):
        """Returns the times of the runs in the region"""
        run_times = []
        for time in range(len(self.region_runs)):
            unformatted_date = self.region_runs[time]["times"]["primary_t"]
            if unformatted_date / 60 < 1:
                seconds = "S:", unformatted_date
                run_times.append(seconds)
            elif unformatted_date / 3600 < 1:
                minutes = "M:", round(unformatted_date / 60, 2)
                run_times.append(minutes)
            else:
                hours = "H:", round(unformatted_date / 3600, 2)
                run_times.append(hours)
        return run_times
    
    @property
    def region_run_platforms(self):
        """Returns the platforms used for runs in the region"""
        final_platforms = []
        for id in range(len(self.region_runs)):
            platform = self.region_runs[id]['system']['platform']
            final_platforms.append(translate_platform(platform))
        return final_platforms

    def __init__(self, *region):
        if len(region) >= 2:
            raise Region_Length_Error("Please only provide one region")
        if len(region[0]) == 0:
            raise Region_Length_Error("Please provide at least one region")
        self.region = region



from speedrunapi import Game_Requests
import speedrunapi.translation as tr
from speedrunapi.errors import URLerror
from datetime import datetime
import urllib.error
import http.client

class Game:
    """Data about a game on speedrun.com
    ARGS: game, the full name of the game on speedrun.com"""

    @property
    def lookup(self):
        """Retruns the full game data from speedrun.com as a dictionary"""
        return self.game_data.game_data

    @property
    def format_name(self):
        """Returns the name of the game on speedrun.com but formated so it can be used in links as a string"""
        self.formated_game_name = ""
        for i in range(len(self.game)):
            if self.game[i] == " ":
                self.formated_game_name = self.formated_game_name + "%20"
            else:
                self.formated_game_name = self.formated_game_name + self.game[i]
        return self.formated_game_name

    @property
    def id(self):
        """Returns the id of the game on speedrun.com as a string"""
        try:
            self.game_id = self.quick_game_data.quick_game_data["data"][0]["id"]
            return self.game_id
        except IndexError:
            return "Game Not Found"

    @property
    def abbreviation(self):
        """Returns the abbreviaion of the name of a game on speedrun.com as a string"""
        return self.game_data.game_data["abbreviation"]

    @property
    def realease_date(self):
        """Returns the data a game was realeased to the public as listed on speedrun.com in a string"""
        return self.game_data.game_data["release-date"]

    @property
    def ruleset(self):
        """Returns the ruleset of a game on speedrun.com as a dictionary"""
        return self.game_data.game_data["ruleset"]

    @property
    def regions(self):
        """Returns the regions of a game on speedrun.com as a dictionary"""
        final_regions = []
        game_regions = self.game_data.game_data["regions"]
        for region in range(len(game_regions)):
            final_regions.append(tr.translate_region(game_regions[region]))
        return final_regions

    @property
    def platforms(self):
        """Returns the platforms of a game on speedrun.com as a dictionary"""
        final_platforms = []
        game_platforms = self.game_data.game_data["platforms"]
        for platform in range(len(game_platforms)):
            final_platforms.append(tr.translate_platform(game_platforms[platform]))
        return final_platforms
    
    @property
    def genres(self):
        """Returns the genres of a game on speedrun.com as a dictionary"""
        final_genres = []
        game_genres = self.game_data.game_data["genres"]
        for genre in range(len(game_genres)):
            final_genres.append(tr.translate_genre(game_genres[genre]))
        return final_genres

    @property
    def engines(self):
        """Returns the engines of a game on speedrun.com as a dictionary"""
        final_engines = []
        game_engines = self.game_data.game_data["engines"]
        for engine in range(len(game_engines)):
            final_engines.append(tr.translate_engine(game_engines[engine]))
        return final_engines

    @property
    def admins(self):
        """Returns the developers, publishers, and moderators of a game on speedrun.com as a dictionary"""
        final_admins = []
        admin_list = ["developers", "publishers", "moderators"]
        for admin in admin_list:
            admins = self.game_data.game_data.get(admin, [])
            if type(admins) == dict:
                admins = next(iter(admins))
            final_admins.append(tr.translate_user_id(admins))
        return final_admins

    @property
    def join_date(self):
        """Returns the data a game joined speedrun.com as a string"""
        try:
            unformatted_date = self.game_data.game_data["created"]
            date_formatted = datetime.fromisoformat(unformatted_date)
            return date_formatted.strftime("%Y-%m-%d")
        except TypeError:
            return "Cannot create join date"
    
    # special cases
    @property
    def catagories(self):
        """Returns the catagories of a game on speedrun.com as a dictionary"""
        self.game_data.game_data_request(3)
        return self.game_data.game_data_response

    @property
    def variables(self):
        """Returns the variables of a game on speedrun.com as a dictionary"""
        self.game_data.game_data_request(4)
        return self.game_data.game_data_response

    @property
    def records(self):
        """Returns the records of a game on speedrun.com as a dictionary"""
        self.game_data.game_data_request(5)
        return self.game_data.game_data_response

    @property
    def series(self):
        """Returns the series of a game on speedrun.com as a dictionary"""
        self.game_data.game_data_request(6)
        return self.game_data.game_data_response

    @property
    def derived_games(self):
        """Returns the derived games of a game on speedrun.com as a dictionary"""
        try:
            self.game_data.game_data_request(7)
            return self.game_data.game_data_response
        except urllib.error.HTTPError:
            return "Derived Games Not Found"
    
    @property
    def romhacks(self):
        """DEPRICATED returns the same as derived games, here for backwards compatiblilty
        Returns the romhacks of a game on speedrun.com as a dictionary"""
        self.game_data.game_data_request(8)
        return self.game_data.game_data_response

    @property
    def leaderboard(self):
        """Returns the leaderboard of a game on speedrun.com as a dictionary"""
        self.game_data.game_data_request(9)
        return self.game_data.game_data_response

    def __init__(self, game):
        try:
            self.game = game
            self.format_name
            self.quick_game_data = Game_Requests(self.formated_game_name)
            self.quick_game_data.game_reqest_quick()
            self.game_id = self.id
            self.game_data = Game_Requests(self.id)
            self.game_data.game_request()
        except http.client.InvalidURL as e:
            details = e.args[0].split("'")[2]
            raise URLerror(f'{details}\nCheck to make sure the name acctualy has no alphanumeric characters/spaces')
        